% Load housing data in CSV into a matlab Table
%% SECTION 1. DATA IMPORT & CLASSIFICATION LEARNER INITIALIZATION %%


T = readtable('housing_regression.xlsx');           % Read file
T.ocean_proximity=categorical(T.ocean_proximity);   % Set variable as categorical

% Choose 5000 rows of data randomly
rng(100) 
idx=randperm(size(T,1));
T1 = T(idx(1:5000),:); 
quant_vals = prctile(T1.median_house_value,[25 50 75]);

% Classification label - high vs low
T1.hi_lo_label =zeros(5000,1);
for ii=1:5000
    if T1.median_house_value(ii) >= quant_vals(3)
        T1.hi_lo_label(ii) =1 ;    %'high'
    end
end

%Train Raw data
rng(201);
[trainInd,~,testInd] = dividerand(size(T1,1),0.7,0,0.3);

Ttest  = T1(testInd',:);
Ttrain = T1(trainInd',:);


classificationLearner(Ttrain,'hi_lo_label'); % Can fit Logistic regression and SVM - both show comparable results



%% Section 2: Visualize the data, Summarize variables, data cleaning, pre-processing if needed
clc;
head(T,10)                                          % show first 10 rows of data table
format compact 
summary(T)                                          % Show the summary of data variables

% From the summary, variable total_bedrooms has 207 missing values.
% Impute missing values with median
jj= isnan(table2array(T(:,5)));
s = summary(T(:,5));
rpl = s.total_bedrooms.Median.*jj;
rpl(rpl==0)= table2array(T(~jj,5));
T(:,5)=table(rpl);

%Only 5 entries with ISLAND. Remove these from data set
jj= find(T.ocean_proximity=='ISLAND');
T(jj,:)=[];

% Reduce data size to 5000 observations to run quickly and demonstrate key
% points in modeling
rng(100) %For reproducibility
idx=randperm(size(T,1));
T1 = T(idx(1:5000),:); 

quant_vals = prctile(T1.median_house_value,[25 50 75]); %Quantiles at 25%,50%,75%

% Classification label - high vs low by $ 264250
T1.hi_lo_label =strings(5000,1);
for ii=1:5000
    if T1.median_house_value(ii) >= quant_vals(3)
        T1.hi_lo_label(ii) ='high';
    else
        T1.hi_lo_label(ii) ='low';
    end
end
%T1.hi_lo_label =categorical(T1.hi_lo_label);

head(T1)

% Visualize the data
nVar =size(T1,2);

figure
for ii =1:nVar-1
subplot(4,3,ii)
xx=table2array(T1(:,ii));
histogram(xx)
b = regexprep(T1.Properties.VariableNames{ii},'_',' ','emptymatch');
title(b)
end

subplot(4,3,nVar)
bar([sum(T1.hi_lo_label=="low") sum(T1.hi_lo_label=="high")] ,'FaceColor',[0 0.4470 0.7410],'FaceAlpha',0.6)
xticklabels({'low','high'})
title('Two group split')
%  Convert categorical variable to nummerical 
op_vbl= size(T1.ocean_proximity);
op_vbl(T1.ocean_proximity =='<1H OCEAN')=1;
op_vbl(T1.ocean_proximity =='INLAND')=2;
op_vbl(T1.ocean_proximity=='NEAR BAY')=3;
op_vbl(T1.ocean_proximity =='NEAR OCEAN')=4;
op_vbl=op_vbl(:);
T1= [T1 table(op_vbl)];

T2 = movevars(T1,'op_vbl','Before','median_house_value');
T1 = T2; % T1 now has a numeric dummy variable for the categorical variable
clear T2;


%% SECTION 3: FIND VARIABLE CORRELATIONS TO EACH OTHER AND THE MEDIAN_HOUSE_VALUE
select_vars =[3 4 8 10 11];
[R,pp] = corr(table2array(T1(:,select_vars)));

kk=0;
figure
for ii = 1:length(select_vars)
    b1 = regexprep(T1.Properties.VariableNames{select_vars(ii)},'_',' ','emptymatch');
    for jj=1:length(select_vars)
        kk=kk+1;
    subplot(5,5,kk)
    xx=table2array(T1(:,[select_vars(ii) select_vars(jj)]));
    plot(xx(:,1),xx(:,2),'.')
    b2 = regexprep(T1.Properties.VariableNames{select_vars(jj)},'_',' ','emptymatch');
    xlabel(b1); ylabel(b2)
    if pp(ii,jj)<0.05
        title(['r = ',num2str(R(ii,jj),2)],'fontsize',12)
    end
    end
end

%% SECTION 4:SPLIT INTO TRAINING AND TEST DATA AND FIT LOGISTIC REGRESSION MODEL
clc; 
% Divide the data set into training, validation and TestIndx
rng(201);
[trainInd,~,testInd] = dividerand(size(T1,1),0.7,0,0.3);

Ttest  = T1(testInd',[1:9 12]);
Ttrain = T1(trainInd',[1:9 12]);

% Logistic Regression model
%X= table2array(Ttrain(:,1:8));
y1 =table2array(Ttrain(:,10));
y=strcmpi('high',string(y1));
mdl = fitglm([Ttrain(:,1:9) table(y)],'Distribution','binomial');


% View the model and the check the predictor coefficients- Ocean proximity
% variables are all insignificant along with total_bedrooms
mdl
%% SECTION 5:DIAGNOSTICS OF MODELS- IDENTIFY OUTLIERS
clc;
% Determine if 95% confidence intervals for the coefficients include 0. If so, you can remove the model terms with those intervals.
confint = coefCI(mdl)
% Remove the redundant predictors and build model again
mdl1 = fitglm([Ttrain(:,[1:4 6:8]) table(y,'variablenames',{'Hi_lo_label'})],'Distribution','binomial');

% Diagnostics of models- identify outliers'
figure;
plotDiagnostics(mdl1,'leverage')
mdl1
%% SECTION 6:TEST MODEL FOR TWO CLASS CLASSIFICATION
y1 =table2array(Ttest(:,10));
yT=strcmpi('high',string(y1));
pred_mdl = predict(mdl1,[Ttest(:,[1:4 6:8]) table(yT,'variablenames',{'Hi_lo_label'})]); %probability values
pred_labels = pred_mdl>0.5;
%Convert labels to strings
pred_labels1=strings;
for ii=1:length(yT)
    if pred_labels(ii)==1
        pred_labels1 (ii)='high';
    else
        pred_labels1 (ii)='low';
    end
end

%Test classification error
misClassf_error =1- mean(pred_labels == yT);

% Confusion Plot

TP = sum((string(Ttest.hi_lo_label)=='high' & pred_labels1'=='high'));
TN = sum((string(Ttest.hi_lo_label)=='low' & pred_labels1'=='low'));
FP = sum((string(Ttest.hi_lo_label)=='low' & pred_labels1'=='high'));
FN = sum((string(Ttest.hi_lo_label)=='high' & pred_labels1'=='low'));
acc = mean(strcmpi(string(Ttest.hi_lo_label),pred_labels1'));



clc;
fprintf(1,'Logistic regression Completed...\n\n')
fprintf(1,'Logistic Regression (7 predictors)\n\t\t\t  Test error = %6.3f\n',misClassf_error)
fprintf(1,'\n\n\n\n')

fprintf(1,['Performance metrics\n\t\t\t  True Positive = %3d\t True Negative = %3d\t', ...
                'False Positive = %3d\t False Negative = %3d\n\n'],TP,TN,FP,FN)

fprintf(1,'\t\t\tAccuracy = %3.3f\t Sensitivity = %3.3f\t Specificity = %3.3f\n',acc,TP/(TP+FN),TN/(TN+FP))
figure;
confusionchart(Ttest.hi_lo_label,(pred_labels1'),'title','2 class Confusion Matrix')

%% SECTION 7: SVM CLASSIFICATION - REGULARIZATION OF VARIABLES DONE AUTOMATICALLY, NO NEED TO CHOOSE FEATURES SEPARATELY AS WAS DONE EARLIER FOR LOGISTIC REGRESSION

SVMModel = fitcsvm(Ttrain(:,1:9),y,'standardize',true);
sv = SVMModel.SupportVectors;
pred_svm = predict(SVMModel,Ttest(:,1:9)); %scores 0 or 1
pred_svm_label(pred_svm==1)="high";
pred_svm_label(pred_svm==0)="low";



%Performance metrics
TP = sum(yT== 1 & pred_svm==1);
TN = sum((string(Ttest.hi_lo_label)=='low' & pred_svm_label'=='low'));
FP = sum((string(Ttest.hi_lo_label)=='low' & pred_svm_label'=='high'));
FN = sum((string(Ttest.hi_lo_label)=='high' & pred_svm_label'=='low'));
acc = mean(strcmpi(string(Ttest.hi_lo_label),pred_svm_label'));
misClassf_error1 = 1-acc;

clc;
fprintf(1,'SVM classification Completed...\n\n')
fprintf(1,'SVM (7 predictors)\n\t\t\t  Test error = %6.3f\n',misClassf_error1)
fprintf(1,'\n\n\n\n')

fprintf(1,['Performance metrics\n\t\t\t  True Positive = %3d\t True Negative = %3d\t', ...
                'False Positive = %3d\t False Negative = %3d\n\n'],TP,TN,FP,FN)

fprintf(1,'\t\t\tAccuracy = %3.3f\t Sensitivity = %3.3f\t Specificity = %3.3f\n',acc,TP/(TP+FN),TN/(TN+FP))
figure
confusionchart(string(Ttest.hi_lo_label),pred_svm_label,'ColumnSummary','column-normalized','RowSummary','row-normalized','Title','SVM - 2 class')

%% SECTION 8:ILLUSTRATE THE SVM BOUNDARY - CHOOSE 2 VARIABLES AND SHOW SVM WITH LINEAR KERNEL VS RBF KERNEL
%Choose 2 variables to illustrate the SVM decision boundary
x1 = table2array(Ttrain(:,3));
x2 = table2array(Ttrain(:,8));
y1 =table2array(Ttrain(:,10));
y=strcmpi('high',string(y1));

%Train the SVM Classifier
cl = fitcsvm([x1 x2],y1);   % Linear SVM

cl2=fitcsvm([x1 x2],y1,'KernelFunction','rbf');


% Predict scores over the grid
d = 0.5;
[x1Grid,x2Grid] = meshgrid(min(x1):d:max(x1),...
    min(x2):d:max(x2));
xGrid = [x1Grid(:),x2Grid(:)];
[~,scores] = predict(cl,xGrid);
[~,scores2] = predict(cl2,xGrid);

figure
gscatter(x1(1:end),x2(1:end),y1(1:end),'rb') % Data is not well separated into classes - for illustrative purposes only
hold on
%h(3) = plot(x1(cl.IsSupportVector(1:10:end)),...
 %   x2(cl.IsSupportVector(1:10:end)),'ko','MarkerSize',10);
    % Support vectors
contour(x1Grid,x2Grid,reshape(scores(:,2),size(x1Grid)),[0 0],'k','linewidth',1.5); %Linear SVM decision boundary
  
contour(x1Grid,x2Grid,reshape(scores2(:,2),size(x1Grid)),[0 0],'m','linewidth',1.5);   % SVM- RBF kernel Decision boundary

xlabel('Age of house');
ylabel('Median Income')
title('Scatter diagram with Decision boundary')
legend("low",'high','Linear SVM','RBF SVM');

%% SECTION 9: SVM MODELFOR MULTICLASS CLASSIFICATION

% Create 3 class labels for low, moderate and high house values
clc;
T1.hi_mod_lo_label =strings(5000,1);
for ii=1:5000
    if T1.median_house_value(ii) >= quant_vals(3)
        T1.hi_mod_lo_label(ii) ='high';
    elseif T1.median_house_value(ii) >= quant_vals(1) && T1.median_house_value(ii) < quant_vals(3)
        T1.hi_mod_lo_label(ii) ='mod';
    elseif T1.median_house_value(ii) < quant_vals(1)
        T1.hi_mod_lo_label(ii) ='low';
    end
end
T1.hi_lo_label =categorical(T1.hi_lo_label);

%Train and Test Data
Ttest  = T1(testInd',[1:8 13]);
Ttrain = T1(trainInd',[1:8 13]);
y = table2array(Ttrain(:,9));

% Multiclass model 
t = templateSVM('Standardize',true,'SaveSupportVectors',true);
predictorNames = Ttrain.Properties.VariableNames(1:8);% {'petalLength','petalWidth'};
responseName = 'hi_mod_lo_label';
classNames = {'low','mod','high'}; % Specify class order
coding = [1 -1 -1; -1 1 -1; -1 -1 1];
Mdl = fitcecoc(Ttrain(:,1:8),y,'Learners',t,'Coding',coding,'ResponseName',responseName,...
    'PredictorNames',predictorNames,'ClassNames',classNames);

L = size(Mdl.CodingMatrix,2); % Number of SVMs
sv = cell(L,1); % Preallocate for support vector indices
for j = 1:L
    SVM = Mdl.BinaryLearners{j};
    sv{j} = SVM.SupportVectors;
    sv{j} = sv{j}.*SVM.Sigma + SVM.Mu;
end
%%

% Test data with One vs rest models
for j=1:L
 scores = predict(Mdl.BinaryLearners{j},Ttest(:,1:8));
 pred_svm= double(scores);

 y= table2array(Ttest(:,9));
 clear yT; yT=NaN(size(y,1),1);
 if j==1
     yT(y=="low")=1;
     yT(y=="mod")=-1;
     yT(y=='high')=-1;
     figtitle{1} = ['Low vs Mod/High'];

 elseif j==2
     yT(y=="low")=-1;
     yT(y=="high")=-1;
       yT(y=='mod')=1;
       figtitle{2} = ['Mod vs Low/High'];

  elseif j==3
    yT(y=="mod")=-1;
    yT(y=="high")=1;
   yT(y=='low')=-1;
    figtitle{3} = ['High vs Low/Mod'];

 end
 yT=yT(:);

TP(j) = sum(yT== 1 & pred_svm ==1);
TN(j) = sum(yT==-1 & pred_svm == -1);
FP(j) = sum(yT==-1 & pred_svm ==1);
FN(j) = sum(yT==1 & pred_svm ==-1);
acc(j) = mean(yT==pred_svm);
misClassf_error1(j) = 1-acc(j);

figure
confusionchart((yT),pred_svm,'title',figtitle{j})
end

clc;
fprintf(1,'\t\tPerformance of One vs All classification- SVM\n')
fprintf(1,'\t%8s\t\t%8s\t\t%8s\t\t%8s\t\t%8s\t\t\n', 'Pos. Class','Neg. Class','Accuracy','Sensitivity','Specificity')
fprintf(1,'\t%s\t\t%10s\t\t\t%0.3f\t\t%0.3f\t\t%0.3f\t\t\r\n','low','High/Mod',acc(1), TP(1)/(TP(1)+FN(1)),TN(1)/(TN(1)+FP(1)));
fprintf(1,'\t%s\t\t%10s\t\t\t%0.3f\t\t%0.3f\t\t%0.3f\t\t\r\n','mod','low/high',acc(2), TP(2)/(TP(2)+FN(2)),TN(2)/(TN(2)+FP(2)));
fprintf(1,'\t%s\t\t%10s\t\t\t%0.3f\t\t%0.3f\t\t%0.3f\t\t\r\n','low','High/Mod',acc(3), TP(3)/(TP(3)+FN(3)),TN(3)/(TN(3)+FP(3)));


% Worst performance when classifying moderately priced houses vs. the other classes. 

%% SECTION 10: MULTICLASS SVM CLASSIFICATION - CHOOSING THE CLASS BASED ON THE PROBABILITY OF BELONGING TO EACH ONE
clc;
%Train and Test Data
Ttest  = T1(testInd',[1:8 13]);
Ttrain = T1(trainInd',[1:8 13]);
y = table2array(Ttrain(:,9));
yT = table2array(Ttest(:,9));


Mdlp = fitcecoc(Ttrain(:,1:8),y,'Learners',t,'FitPosterior',true,...
    'ClassNames',{'low','mod','high'},...
    'Verbose',2);

% Predict training sample
[label,~,~,Posterior] = resubPredict(Mdlp,'Verbose',1);
Mdlp.BinaryLoss

% Display random set of results
idx = randsample(size(Ttrain,1),10,1);
Mdl.ClassNames
clc
table(y(idx),label(idx),Posterior(idx,:),...
    'VariableNames',{'TrueLabel','PredLabel','Posterior'})

% Predict on test data
[label,~,~,Posterior] = predict(Mdlp,Ttest(:,1:8),'Verbose',1);
idx = randsample(size(Ttest,1),10,1);
Mdl.ClassNames

table(yT(idx),label(idx),Posterior(idx,:),...
    'VariableNames',{'TrueLabel','PredLabel','Posterior'})

figure
confusionchart(yT,string(label),'title','3 class SVM classification','RowSummary','row-normalized')

